@@ -1,16 +1,27 @@
 'use strict';
 
-const binding = process.binding('buffer');
+const { Math } = primordials;
+
 const {
   ERR_BUFFER_OUT_OF_BOUNDS,
+  ERR_INVALID_ARG_TYPE,
   ERR_OUT_OF_RANGE
 } = require('internal/errors').codes;
 const { validateNumber } = require('internal/validators');
-const { setupBufferJS } = binding;
-
-// Remove from the binding so that function is only available as exported here.
-// (That is, for internal use only.)
-delete binding.setupBufferJS;
+const {
+  asciiSlice,
+  base64Slice,
+  latin1Slice,
+  hexSlice,
+  ucs2Slice,
+  utf8Slice,
+  asciiWrite,
+  base64Write,
+  latin1Write,
+  hexWrite,
+  ucs2Write,
+  utf8Write
+} = internalBinding('buffer');
 
 // Temporary buffers to convert numbers.
 const float32Array = new Float32Array(1);
@@ -105,9 +116,9 @@ function readBigInt64LE(offset = 0) {
     (last << 24); // Overflow
   return (BigInt(val) << 32n) +
     BigInt(first +
-      this[++offset] * 2 ** 8 +
-      this[++offset] * 2 ** 16 +
-      this[++offset] * 2 ** 24);
+    this[++offset] * 2 ** 8 +
+    this[++offset] * 2 ** 16 +
+    this[++offset] * 2 ** 24);
 }
 
 function readBigInt64BE(offset = 0) {
@@ -123,12 +134,14 @@ function readBigInt64BE(offset = 0) {
     this[++offset];
   return (BigInt(val) << 32n) +
     BigInt(this[++offset] * 2 ** 24 +
-      this[++offset] * 2 ** 16 +
-      this[++offset] * 2 ** 8 +
-      last);
+    this[++offset] * 2 ** 16 +
+    this[++offset] * 2 ** 8 +
+    last);
 }
 
 function readUIntLE(offset, byteLength) {
+  if (offset === undefined)
+    throw new ERR_INVALID_ARG_TYPE('offset', 'number', offset);
   if (byteLength === 6)
     return readUInt48LE(this, offset);
   if (byteLength === 5)
@@ -139,7 +152,7 @@ function readUIntLE(offset, byteLength) {
     return this.readUInt32LE(offset);
   if (byteLength === 2)
     return this.readUInt16LE(offset);
-  if (byteLength === 1 || byteLength === undefined)
+  if (byteLength === 1)
     return this.readUInt8(offset);
 
   boundsError(byteLength, 6, 'byteLength');
@@ -216,6 +229,8 @@ function readUInt8(offset = 0) {
 }
 
 function readUIntBE(offset, byteLength) {
+  if (offset === undefined)
+    throw new ERR_INVALID_ARG_TYPE('offset', 'number', offset);
   if (byteLength === 6)
     return readUInt48BE(this, offset);
   if (byteLength === 5)
@@ -226,7 +241,7 @@ function readUIntBE(offset, byteLength) {
     return this.readUInt32BE(offset);
   if (byteLength === 2)
     return this.readUInt16BE(offset);
-  if (byteLength === 1 || byteLength === undefined)
+  if (byteLength === 1)
     return this.readUInt8(offset);
 
   boundsError(byteLength, 6, 'byteLength');
@@ -294,6 +309,8 @@ function readUInt16BE(offset = 0) {
 }
 
 function readIntLE(offset, byteLength) {
+  if (offset === undefined)
+    throw new ERR_INVALID_ARG_TYPE('offset', 'number', offset);
   if (byteLength === 6)
     return readInt48LE(this, offset);
   if (byteLength === 5)
@@ -304,7 +321,7 @@ function readIntLE(offset, byteLength) {
     return this.readInt32LE(offset);
   if (byteLength === 2)
     return this.readInt16LE(offset);
-  if (byteLength === 1 || byteLength === undefined)
+  if (byteLength === 1)
     return this.readInt8(offset);
 
   boundsError(byteLength, 6, 'byteLength');
@@ -384,6 +401,8 @@ function readInt8(offset = 0) {
 }
 
 function readIntBE(offset, byteLength) {
+  if (offset === undefined)
+    throw new ERR_INVALID_ARG_TYPE('offset', 'number', offset);
   if (byteLength === 6)
     return readInt48BE(this, offset);
   if (byteLength === 5)
@@ -394,7 +413,7 @@ function readIntBE(offset, byteLength) {
     return this.readInt32BE(offset);
   if (byteLength === 2)
     return this.readInt16BE(offset);
-  if (byteLength === 1 || byteLength === undefined)
+  if (byteLength === 1)
     return this.readInt8(offset);
 
   boundsError(byteLength, 6, 'byteLength');
@@ -592,7 +611,7 @@ function writeBigInt64BE(value, offset = 0) {
     this, value, offset, -0x8000000000000000n, 0x7fffffffffffffffn);
 }
 
-function writeUIntLE(value, offset = 0, byteLength) {
+function writeUIntLE(value, offset, byteLength) {
   if (byteLength === 6)
     return writeU_Int48LE(this, value, offset, 0, 0xffffffffffff);
   if (byteLength === 5)
@@ -603,7 +622,7 @@ function writeUIntLE(value, offset = 0, byteLength) {
     return writeU_Int32LE(this, value, offset, 0, 0xffffffff);
   if (byteLength === 2)
     return writeU_Int16LE(this, value, offset, 0, 0xffff);
-  if (byteLength === 1 || byteLength === undefined)
+  if (byteLength === 1)
     return writeU_Int8(this, value, offset, 0, 0xff);
 
   boundsError(byteLength, 6, 'byteLength');
@@ -703,7 +722,7 @@ function writeUInt8(value, offset = 0) {
   return writeU_Int8(this, value, offset, 0, 0xff);
 }
 
-function writeUIntBE(value, offset = 0, byteLength) {
+function writeUIntBE(value, offset, byteLength) {
   if (byteLength === 6)
     return writeU_Int48BE(this, value, offset, 0, 0xffffffffffffff);
   if (byteLength === 5)
@@ -714,7 +733,7 @@ function writeUIntBE(value, offset = 0, byteLength) {
     return writeU_Int32BE(this, value, offset, 0, 0xffffffff);
   if (byteLength === 2)
     return writeU_Int16BE(this, value, offset, 0, 0xffff);
-  if (byteLength === 1 || byteLength === undefined)
+  if (byteLength === 1)
     return writeU_Int8(this, value, offset, 0, 0xff);
 
   boundsError(byteLength, 6, 'byteLength');
@@ -795,7 +814,7 @@ function writeUInt16BE(value, offset = 0) {
   return writeU_Int16BE(this, value, offset, 0, 0xffff);
 }
 
-function writeIntLE(value, offset = 0, byteLength) {
+function writeIntLE(value, offset, byteLength) {
   if (byteLength === 6)
     return writeU_Int48LE(this, value, offset, -0x800000000000, 0x7fffffffffff);
   if (byteLength === 5)
@@ -806,7 +825,7 @@ function writeIntLE(value, offset = 0, byteLength) {
     return writeU_Int32LE(this, value, offset, -0x80000000, 0x7fffffff);
   if (byteLength === 2)
     return writeU_Int16LE(this, value, offset, -0x8000, 0x7fff);
-  if (byteLength === 1 || byteLength === undefined)
+  if (byteLength === 1)
     return writeU_Int8(this, value, offset, -0x80, 0x7f);
 
   boundsError(byteLength, 6, 'byteLength');
@@ -824,7 +843,7 @@ function writeInt8(value, offset = 0) {
   return writeU_Int8(this, value, offset, -0x80, 0x7f);
 }
 
-function writeIntBE(value, offset = 0, byteLength) {
+function writeIntBE(value, offset, byteLength) {
   if (byteLength === 6)
     return writeU_Int48BE(this, value, offset, -0x800000000000, 0x7fffffffffff);
   if (byteLength === 5)
@@ -835,7 +854,7 @@ function writeIntBE(value, offset = 0, byteLength) {
     return writeU_Int32BE(this, value, offset, -0x80000000, 0x7fffffff);
   if (byteLength === 2)
     return writeU_Int16BE(this, value, offset, -0x8000, 0x7fff);
-  if (byteLength === 1 || byteLength === undefined)
+  if (byteLength === 1)
     return writeU_Int8(this, value, offset, -0x80, 0x7f);
 
   boundsError(byteLength, 6, 'byteLength');
@@ -906,54 +925,72 @@ function writeFloatBackwards(val, offset = 0) {
   return offset;
 }
 
-// FastBuffer wil be inserted here by lib/buffer.js
+class FastBuffer extends Uint8Array {}
+
+function addBufferPrototypeMethods(proto) {
+  proto.readBigUInt64LE = readBigUInt64LE,
+  proto.readBigUInt64BE = readBigUInt64BE,
+  proto.readBigInt64LE = readBigInt64LE,
+  proto.readBigInt64BE = readBigInt64BE,
+  proto.writeBigUInt64LE = writeBigUInt64LE,
+  proto.writeBigUInt64BE = writeBigUInt64BE,
+  proto.writeBigInt64LE = writeBigInt64LE,
+  proto.writeBigInt64BE = writeBigInt64BE,
+
+  proto.readUIntLE = readUIntLE;
+  proto.readUInt32LE = readUInt32LE;
+  proto.readUInt16LE = readUInt16LE;
+  proto.readUInt8 = readUInt8;
+  proto.readUIntBE = readUIntBE;
+  proto.readUInt32BE = readUInt32BE;
+  proto.readUInt16BE = readUInt16BE;
+  proto.readIntLE = readIntLE;
+  proto.readInt32LE = readInt32LE;
+  proto.readInt16LE = readInt16LE;
+  proto.readInt8 = readInt8;
+  proto.readIntBE = readIntBE;
+  proto.readInt32BE = readInt32BE;
+  proto.readInt16BE = readInt16BE;
+
+  proto.writeUIntLE = writeUIntLE;
+  proto.writeUInt32LE = writeUInt32LE;
+  proto.writeUInt16LE = writeUInt16LE;
+  proto.writeUInt8 = writeUInt8;
+  proto.writeUIntBE = writeUIntBE;
+  proto.writeUInt32BE = writeUInt32BE;
+  proto.writeUInt16BE = writeUInt16BE;
+  proto.writeIntLE = writeIntLE;
+  proto.writeInt32LE = writeInt32LE;
+  proto.writeInt16LE = writeInt16LE;
+  proto.writeInt8 = writeInt8;
+  proto.writeIntBE = writeIntBE;
+  proto.writeInt32BE = writeInt32BE;
+  proto.writeInt16BE = writeInt16BE;
+
+  proto.readFloatLE = bigEndian ? readFloatBackwards : readFloatForwards;
+  proto.readFloatBE = bigEndian ? readFloatForwards : readFloatBackwards;
+  proto.readDoubleLE = bigEndian ? readDoubleBackwards : readDoubleForwards;
+  proto.readDoubleBE = bigEndian ? readDoubleForwards : readDoubleBackwards;
+  proto.writeFloatLE = bigEndian ? writeFloatBackwards : writeFloatForwards;
+  proto.writeFloatBE = bigEndian ? writeFloatForwards : writeFloatBackwards;
+  proto.writeDoubleLE = bigEndian ? writeDoubleBackwards : writeDoubleForwards;
+  proto.writeDoubleBE = bigEndian ? writeDoubleForwards : writeDoubleBackwards;
+
+  proto.asciiSlice = asciiSlice;
+  proto.base64Slice = base64Slice;
+  proto.latin1Slice = latin1Slice;
+  proto.hexSlice = hexSlice;
+  proto.ucs2Slice = ucs2Slice;
+  proto.utf8Slice = utf8Slice;
+  proto.asciiWrite = asciiWrite;
+  proto.base64Write = base64Write;
+  proto.latin1Write = latin1Write;
+  proto.hexWrite = hexWrite;
+  proto.ucs2Write = ucs2Write;
+  proto.utf8Write = utf8Write;
+}
+
 module.exports = {
-  setupBufferJS,
-  // Container to export all read write functions.
-  readWrites: {
-    readBigUInt64LE,
-    readBigUInt64BE,
-    readBigInt64LE,
-    readBigInt64BE,
-    writeBigUInt64LE,
-    writeBigUInt64BE,
-    writeBigInt64LE,
-    writeBigInt64BE,
-    readUIntLE,
-    readUInt32LE,
-    readUInt16LE,
-    readUInt8,
-    readUIntBE,
-    readUInt32BE,
-    readUInt16BE,
-    readIntLE,
-    readInt32LE,
-    readInt16LE,
-    readInt8,
-    readIntBE,
-    readInt32BE,
-    readInt16BE,
-    writeUIntLE,
-    writeUInt32LE,
-    writeUInt16LE,
-    writeUInt8,
-    writeUIntBE,
-    writeUInt32BE,
-    writeUInt16BE,
-    writeIntLE,
-    writeInt32LE,
-    writeInt16LE,
-    writeInt8,
-    writeIntBE,
-    writeInt32BE,
-    writeInt16BE,
-    readFloatLE: bigEndian ? readFloatBackwards : readFloatForwards,
-    readFloatBE: bigEndian ? readFloatForwards : readFloatBackwards,
-    readDoubleLE: bigEndian ? readDoubleBackwards : readDoubleForwards,
-    readDoubleBE: bigEndian ? readDoubleForwards : readDoubleBackwards,
-    writeFloatLE: bigEndian ? writeFloatBackwards : writeFloatForwards,
-    writeFloatBE: bigEndian ? writeFloatForwards : writeFloatBackwards,
-    writeDoubleLE: bigEndian ? writeDoubleBackwards : writeDoubleForwards,
-    writeDoubleBE: bigEndian ? writeDoubleForwards : writeDoubleBackwards
-  }
+  FastBuffer,
+  addBufferPrototypeMethods
 };
