@@ -1,16 +1,15 @@
 'use strict';
 
-const { ModuleWrap, callbackMap } = internalBinding('module_wrap');
-const debug = require('util').debuglog('esm');
-const ArrayJoin = Function.call.bind(Array.prototype.join);
-const ArrayMap = Function.call.bind(Array.prototype.map);
+const { ArrayPrototype } = primordials;
+
+const debug = require('internal/util/debuglog').debuglog('esm');
 
 const createDynamicModule = (exports, url = '', evaluate) => {
   debug('creating ESM facade for %s with exports: %j', url, exports);
-  const names = ArrayMap(exports, (name) => `${name}`);
+  const names = ArrayPrototype.map(exports, (name) => `${name}`);
 
   const source = `
-${ArrayJoin(ArrayMap(names, (name) =>
+${ArrayPrototype.join(ArrayPrototype.map(names, (name) =>
     `let $${name};
 export { $${name} as ${name} };
 import.meta.exports.${name} = {
@@ -21,7 +20,7 @@ import.meta.exports.${name} = {
 
 import.meta.done();
 `;
-
+  const { ModuleWrap, callbackMap } = internalBinding('module_wrap');
   const m = new ModuleWrap(source, `${url}`);
   m.link(() => 0);
   m.instantiate();
